# Copyright 2023 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import json
import os
import time
from glob import glob
from pathlib import Path
from threading import Thread
from typing import Optional

import nbformat
from nbconvert.preprocessors.execute import ExecutePreprocessor
from nbformat import NotebookNode
from sapientml.params import CancellationToken
from sapientml.util.logging import setup_logger

from .code_template import Code_Template
from .pipeline_explanation import Pipeline_Explanation

# To surpress warinings when executing ipynb
os.environ["PYDEVD_DISABLE_FILE_VALIDATION"] = "1"


class AST_Update:
    def __init__(self, visualization_code=None, logger=None, local_path=None):
        self.visualization_code = visualization_code
        self.logger = logger
        self.local_path = local_path
        if logger is None:
            self.logger = setup_logger()

        self.FUNCION_MAPPING = {
            "_STRING_COLS_WITH_MISSING_VALUES": "__cat_missing_data__",
            "_NUMERIC_COLS_WITH_MISSING_VALUES": "__num_missing_data__",
            "_NUMERIC_COLS_TO_SCALE =": "__num_to_scale__",
            "# SUBMISSION": "__set_submission__",
            "__feature_train = __train_dataset": "__set_target__",
            "__model.fit": "__fit__",
            "_ohe = OrdinalEncoder": "__OrdinalEncoder_exp__",
            "__train_dataset.drop([": "__drop_Irrelevant__",
            "def process_text": "__process_text_exp__",
            "__tfidfvectorizer =": "__tfidfvectorizer_exp__",
            "__y_pred = __model.predict_proba": "__predict_proba_exp__",
            "smote = SMOTE()": "__smote__exp__",
            "__standard_scaler = StandardScaler()": "__standard_scaler__exp__",
            "__simple_imputer = SimpleImputer(": "__simple_imputer__exp__",
            # The followings all mean "after #LOAD DATA block"
            "# TIME SERIES SETTING AND GENERATING FEATURES": "__test_dataset_submission_columns__exp__",
            "# TRAIN-TEST SPLIT": "__test_dataset_submission_columns__exp__",
            "# HANDLE MIXED TYPE": "__test_dataset_submission_columns__exp__",
            "# CONVERT INF TO NAN": "__test_dataset_submission_columns__exp__",
            "# HANDLE JAPANESE TEXT": "__test_dataset_submission_columns__exp__",
            "# HANDLE ITERABLE VALUES IN DATAFRAME": "__test_dataset_submission_columns__exp__",
            "# STORE SUBMISSION RELEVANT COLUMNS": "__test_dataset_submission_columns__exp__",
            "# PREPROCESSING-1": "__test_dataset_submission_columns__exp__",
            "# DETACH TARGET": "__test_dataset_submission_columns__exp__",
        }

        # add functionname:Bool to show the function can be added as duplicate
        self.FUNCION_MAPPING_DUPLICATE_FLAG = {
            "__standard_scaler__exp__": False,
            "__simple_imputer__exp__": False,
            "__smote__exp__": False,
            "__test_dataset_submission_columns__exp__": False,
        }
        self.last_status = {}

    def process(self, loc, prev):
        for func in self.FUNCION_MAPPING:
            # run loc against all function calls
            # if there is a match, the it will be executed
            if loc.strip().startswith(func):
                try:
                    execute = False
                    fired_func = self.FUNCION_MAPPING[func]
                    if not self.FUNCION_MAPPING_DUPLICATE_FLAG.get(fired_func, True):
                        # there is flag for fired_function
                        if not self.last_status.get(fired_func, False):
                            # previously the flag didn't changed
                            self.last_status[fired_func] = True
                            execute = True
                    else:
                        # there isn't any flag for fired_function
                        execute = True

                    if execute:
                        return self.__getattribute__(fired_func)(loc.strip(), prev)
                except AttributeError:
                    raise NotImplementedError(f"{fired_func}")

    def __fit__(self, loc, prev):
        added_codes = []
        added_codes.append(
            (
                [
                    "# Training and Prediction",
                    "First, we will train a model based on the pre-processed training dataset.",
                    "Second, let's predict test values based on the trained model.",
                ],
                "markdown",
            )
        )
        if "__model = CatBoostRegressor()" in prev:
            added_codes.append(
                (
                    [
                        "## CatBoost Regression",
                        "We will use *CatBoostRegressor* which is a fast, scalable, high performance gradient boosting on decision trees library. Used for ranking, classification, regression and other ML tasks.",
                        "More details about *CatBoostRegressor* can be found [here](https://catboost.ai/docs/installation/python-installation-method-pip-install).",
                    ],
                    "markdown",
                )
            )
        elif "__model = RandomForestClassifier()" in prev:
            added_codes.append(
                (
                    [
                        "## Random Forest Classifier",
                        "We will use *RandomForestClassifier* which is a meta estimator that fits a number of decision tree classifiers on various sub-samples of the dataset and uses averaging to improve the predictive accuracy and control over-fitting.",
                        "The sub-sample size is controlled with the max_samples parameter if bootstrap=True (default), otherwise the whole dataset is used to build each tree.",
                        "More details about *RandomForestClassifier* can be found [here](https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html).",
                    ],
                    "markdown",
                )
            )
        elif "__model = CatBoostClassifier()" in prev:
            added_codes.append(
                (
                    [
                        "## Cat Boost Classifier",
                        "We will use *CatBoostClassifier* which is Training and applying models for the classification problems. It provides compatibility with the scikit-learn tools.",
                        "More details about *CatBoostClassifier* can be found [here](https://catboost.ai/en/docs/concepts/python-reference_catboostclassifier).",
                    ],
                    "markdown",
                )
            )
        elif "__model = LGBMRegressor()" in prev:
            added_codes.append(
                (
                    [
                        "## LightGBM Regressor",
                        "We will use *LightGBM Regressor* which is constructing a gradient boosting model. We will use *lightgbm* package.",
                        "More details about *LightGBM Regressor* can be found [here](https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMRegressor.html).",
                    ],
                    "markdown",
                )
            )
        elif "__model = XGBClassifier()" in prev:
            added_codes.append(
                (
                    [
                        "## XGBClassifier",
                        "We will use *XGBClassifier* from XGBoost package which is used for supervised learning problems, where we use the training data (with multiple features) to predict a target variable.",
                        "More details about *XGBClassifier* can be found [here](https://xgboost.readthedocs.io/en/stable/tutorials/model.html).",
                    ],
                    "markdown",
                )
            )
        elif "__model = LogisticRegression()" in prev:
            added_codes.append(
                (
                    [
                        "## LogisticRegression",
                        """Logistic regression, despite its name, is a linear model for classification rather than regression. Logistic regression is also known in the literature as logit regression, maximum-entropy classification (MaxEnt) or the log-linear classifier.
                                 In this model, the probabilities describing the possible outcomes of a single trial are modeled using a logistic function.""",
                        "More details about *LogisticRegression* can be found [here](https://scikit-learn.org/stable/modules/linear_model.html#logistic-regression).",
                    ],
                    "markdown",
                )
            )
        elif "__model = ExtraTreesRegressor()" in prev:
            added_codes.append(
                (
                    [
                        "## ExtraTreesRegressor",
                        """The SK-Learn.ensemble module includes two averaging algorithms based on randomized decision trees:
                                     the RandomForest algorithm and the Extra-Trees method. Both algorithms are perturb-and-combine techniques specifically designed for trees.
                                     This means a diverse set of classifiers is created by introducing randomness in the classifier construction.
                                     The prediction of the ensemble is given as the averaged prediction of the individual classifiers.
                                     In extremely randomized trees (*ExtraTreesRegressor*), randomness goes one step further in the way splits are computed.
                                     As in random forests, a random subset of candidate features is used, but instead of looking for the most discriminative thresholds, thresholds are drawn at random for each candidate feature and the best of these randomly-generated thresholds is picked as the splitting rule. """,
                        "More details about *ExtraTreesRegressor* can be found [here](https://scikit-learn.org/stable/modules/ensemble.html#forest).",
                    ],
                    "markdown",
                )
            )
        elif "__model = LinearRegression()" in prev:
            added_codes.append(
                (
                    [
                        "## LinearRegression",
                        """LinearRegression fits a linear model with coefficients w = (w1, …, wp) to minimize the residual sum of squares between the observed targets in the dataset, and the targets predicted by the linear approximation.""",
                        "More details about *LinearRegression* can be found [here](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html).",
                    ],
                    "markdown",
                )
            )
        elif "__model = DecisionTreeClassifier()" in prev:
            added_codes.append(
                (
                    [
                        "## Decision Tree Classifier",
                        """Decision Trees (DTs) are a non-parametric supervised learning method used for classification and regression.
                                 The goal is to create a model that predicts the value of a target variable by learning simple decision rules inferred from the data features.
                                 A tree can be seen as a piecewise constant approximation.""",
                        "More details about *DecisionTreeClassifier* can be found [here](https://scikit-learn.org/stable/modules/tree.html#tree).",
                    ],
                    "markdown",
                )
            )
        elif "__model = XGBRegressor()" in prev:
            added_codes.append(
                (
                    [
                        "## XGBRegressor",
                        """ We will use *XGBRegressor* which is "Extreme Gradient Boosting" approach and it is an implementation of gradient boosting trees algorithm.
                                 The XGBoost is a popular supervised machine learning model with characteristics like computation speed, parallelization, and performance.""",
                        "More details about *XGBRegressor* can be found [here](https://xgboost.readthedocs.io/en/latest/index.html).",
                    ],
                    "markdown",
                )
            )

        return added_codes

    def __update_dataset_path__(self, loc, prev):
        if self.local_path:
            return f"dataset_path = {self.local_path}"

    def __set_submission__(self, loc, prev):
        added_codes = []
        added_codes.append(
            (
                [
                    "# Submission File",
                    'We have to maintain the target columns in "submission.csv" which will be submitted as our prediction results.',
                ],
                "markdown",
            )
        )
        return added_codes

    def __simple_imputer__exp__(self, loc, prev):
        added_codes = []
        added_codes.append(
            (
                [
                    "## Imputation Transformer",
                    "We will use *SimpleImputer* which is an imputation transformer for completing missing values.",
                    "We can use out-of-the-box imputation transformer from Scikit-Learn packages. The detail and the list of complete parameters can be found [here](https://scikit-learn.org/stable/modules/generated/sklearn.impute.SimpleImputer.html).",
                ],
                "markdown",
            )
        )
        return added_codes

    def __standard_scaler__exp__(self, loc, prev):
        added_codes = []
        added_codes.append(
            (
                [
                    "## Standard Scaler"
                    "\nWe will use Scikit-Learn  *StandardScaler* which standardizes features by removing the mean and scaling to unit variance.",
                    "The deatil can be found [here](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html).",
                ],
                "markdown",
            )
        )
        return added_codes

    def __predict_proba_exp__(self, loc, prev):
        added_codes = []
        added_codes.append(
            (
                [
                    "We need to predict the probabilities for each group, therefore we will use **predict_proba()** features that generates probability for each group.",
                ],
                "markdown",
            )
        )
        return added_codes

    def __smote__exp__(self, loc, prev):
        added_codes = []
        added_codes.append(
            (
                [
                    "## SMOTE"
                    "\nWe will use SMOTE (Synthetic Minority Over-sampling Technique) as presented by Chawla et al. (the paper is available [here](https://www.jair.org/index.php/jair/article/view/10302).)",
                    "Implementation and examples of SMOTE is available in [Imbalanced-Learn](https://imbalanced-learn.org/stable/references/generated/imblearn.over_sampling.SMOTE.html#r001eabbe5dd7-1).",
                ],
                "markdown",
            )
        )
        return added_codes

    def __process_text_exp__(self, loc, prev):
        added_codes = []
        if "_TEXT_COLUMNS" in prev:
            tree = ast.parse(prev)
            keys = [_.value for _ in tree.body[0].value.elts]
            singular = "s" if len(keys) > 1 else ""
            if len(keys):
                added_codes.append(
                    (
                        [
                            "# Text Processing",
                            f"The dataset has **{len(keys)}** text value{singular} as follows: **{u','.join(_ for _ in keys)}**.",
                            "Now, let's covert the text as follows.\n",
                            "- First, convert text to lowercase;\n",
                            "- Second, strip all punctuations;\n",
                            # TODO: Does this rule apply always?!
                            "- Finally, convert all numbers in text to 'num'; therefore, in the next step our model will use a single token instead of valriety of tokens of numbers.",
                        ],
                        "markdown",
                    )
                )
        return added_codes

    def __tfidfvectorizer_exp__(self, loc, prev):
        added_codes = []
        added_codes.append(
            (
                [
                    "# Text Vectorizer",
                    "In the next step, we will transfer pre-processed text columns to a vector representation. The vector representations allows us to train a model based on numerical representations.",
                    "We will use TfidfVectorizer and more detail can be found [here](https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html).",
                ],
                "markdown",
            )
        )
        return added_codes

    def __drop_Irrelevant__(self, loc, prev):
        added_codes = []
        try:
            if "DISCARD IRRELEVANT COLUMNS" in prev:
                tree = ast.parse(loc)
                keys = [_.value for _ in tree.body[0].value.args[0].elts]
                plural = "s" if len(keys) > 1 else ""
                added_codes.append(
                    (
                        [
                            "### Discard Irrelevant Columns",
                            f'In the given input dataset there are **{len(keys)}** column{plural} that can be removed as follows: {", ".join(_ for _ in  keys)}.',
                        ],
                        "markdown",
                    )
                )
        except Exception as ex:
            self.logger.warning(f"Warning 134: target column extraction issue ({ex})")
            return []
        return added_codes

    def __OrdinalEncoder_exp__(self, loc, prev):
        added_codes = []
        try:
            if "_CATEGORICAL_COLS" in prev:
                tree = ast.parse(prev)
                keys = [_.value for _ in tree.body[0].value.elts]
                plural = "s" if len(keys) > 1 else ""
                added_codes.append(
                    (
                        [
                            "## Encoding Ordinal Categorical Features",
                            "Let's transfer categorical features as an integer array.",
                            "We will use Ordinal Encoder as explained [here](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.OrdinalEncoder.html).",
                            f'\nIn the given input dataset there are **{len(keys)}** column{plural} that can be transfered to integer and it includes:* {", ".join(_ for _ in  keys)} *.',
                        ],
                        "markdown",
                    )
                )
        except Exception as ex:
            self.logger.warning(f"Warning 134: target column extraction issue ({ex})")
            return []
        return added_codes

    def __set_submission_cols__(self, loc, prev):
        added_codes = []
        try:
            tree = ast.parse(loc)
            cols_elts = tree.body[0].value.args[0].elts
            cols = [_.value for _ in cols_elts]
            added_codes.append(
                (
                    [
                        "# Generate Submission File",
                        "We have to maintain the following columns in *prediction_result.csv*.",
                        "then, we can drop that column(s) from original dataset because it is unique and not useful for training a model.",
                        "In some cases, an ID may carry useful information such as student ID where it may consist of admission year and other related info.",
                    ],
                    "markdown",
                )
            )
            if len(cols):
                added_codes.append(([f"""submission_columns = ['{u"','".join(_ for _ in cols)}']"""], "code"))
                added_codes.append((["submission=pd.DataFrame(__test_dataset[submission_columns].copy())"], "code"))
            else:
                self.logger.warning("Warning 608: submission file cannot be generated.")
                added_codes.append((["submission_columns = []"], "code"))

        except Exception as ex:
            self.logger.warning(f"Warning 134: target column extraction issue ({ex})")
            return []
        return added_codes

    def __set_target__(self, loc, prev):
        added_codes = []
        try:
            tree = ast.parse(loc)
            target_col = [_.value for _ in tree.body[0].value.args[0].elts]
            singular = "s" if len(target_col) > 1 else ""
            singular_verb = "are" if len(target_col) > 1 else "is"
            added_codes.append(
                (
                    [
                        f"### Target Column{singular}",
                        f"We need to predict the target column{singular}.",
                        f"Therefore, we need to detach the target column{singular} in prediction.",
                        f"Note that if we don't drop the field{singular}, it will generate a model with high accuracy on training and worst accuracy on test (because the value in test dataset is Null).",
                        f'Here {singular_verb} the list of *target column{singular}*: **{u", ".join(_ for _ in target_col)}**',
                    ],
                    "markdown",
                )
            )
        except Exception as ex:
            self.logger.warning(f"Warning 134: target column extraction issue ({ex})")
            return []
        except Exception as ex:
            self.logger.warning(f"Warning 134: visualization processing issue ({ex})")
            return added_codes
        return added_codes

    def __dataset_drop__(self, loc, prev):
        description = []
        keys = []
        tree = ast.parse(loc)
        if tree:
            if tree.body:
                if isinstance(tree.body[0], ast.Assign):
                    if tree.body[0].targets[0].id == "__feature":  # target col
                        for key in tree.body[0].value.args[0].elts:
                            keys.append(key.value)
        singular = "s" if len(keys) > 1 else ""
        # singular_verb = "are" if len(keys) > 1 else "is"
        description.append(
            (
                [
                    f"## Drop Target Column{singular}",
                    f"We need to drop target column{singular} from the training dataset.",
                    f"Now let's drops the target column{singular} of **{u','.join(_ for _ in keys)}** from the dataset.",
                ],
                "markdown",
            )
        )
        return description

    def __num_missing_data__(self, loc, prev):
        description = []
        keys = []
        tree = ast.parse(loc)
        if tree:
            if tree.body:
                if isinstance(tree.body[0], ast.Assign):
                    for key in tree.body[0].value.elts:
                        keys.append(key.value)
        recs = ", ".join(_ for _ in keys)
        singular = "s" if len(keys) > 1 else ""
        singular_verb = "are" if len(keys) > 1 else "is"
        description.append(([f"## Remove Missing Values in Numerical Column{singular}"], "markdown"))
        description.append(
            (
                [
                    f"In the given input dataset there {singular_verb} **{len(keys)} column{singular}** with missing data as follows: {recs}.",
                    f"\nThe following code removes the missing values from those column{singular}. We use an average value of each column or 0 to replace the Null values according to the missing value ratio.",
                ],
                "markdown",
            )
        )
        return description

    def __cat_missing_data__(self, loc, prev):
        description = []
        keys = []
        tree = ast.parse(loc)
        if tree:
            if tree.body:
                if isinstance(tree.body[0], ast.Assign):
                    for key in tree.body[0].value.elts:
                        keys.append(key.value)
        recs = ", ".join(_ for _ in keys)
        description.append(
            (
                [
                    "## Remove Missing Values in Categorical Columns",
                    f'\nIn the given input dataset there are **{len(keys)} column{"s" if len(keys)>1 else ""}** with missing data as follows: {recs}.',
                    "\nThe following code removes the missing values from those columns. We use the most frequent value of each column or empty character to replace the Null values according to the missing value ratio.",
                ],
                "markdown",
            )
        )
        return description

    def __ordinal_encoder__(self, loc, prev):
        description = []
        keys = []
        tree = ast.parse(loc)
        if tree:
            if tree.body:
                if isinstance(tree.body[0], ast.Assign):
                    keys = [_.value for _ in tree.body[0].value.elts]
                    description.append(
                        (
                            [
                                "## Encoding Ordinal Categorical Features",
                                "We will encode categorical features as an integer array.",
                                f'In the given input dataset there are *{len(keys)}* column{"s" if len(keys)>1 else ""} with string values as follows: {(u", ".join(_) for _ in keys)}',
                            ],
                            "markdown",
                        )
                    )

        description.append((["The following code encode the selected columns."], "markdown"))
        return description

    def __num_to_scale__(self, loc, prev):
        description = []
        keys = []
        tree = ast.parse(loc)
        if tree:
            if tree.body:
                if isinstance(tree.body[0], ast.Assign):
                    keys = [_.value for _ in tree.body[0].value.elts]
        description.append((["## Numeric to Scale"], "markdown"))
        description.append(
            (
                [
                    f'In the given input dataset there are **{len(keys)} column{"s" if len(keys)>1 else ""}** with numeric values as follows where we can convert those values to scale through [log1p](https://numpy.org/doc/stable/reference/generated/numpy.log1p.html).'
                ],
                "markdown",
            )
        )
        recs = ", ".join(_ for _ in keys)
        description.append(([recs], "markdown"))
        return description

    def __test_dataset_submission_columns__exp__(self, loc, prev):
        added_codes = []
        try:
            if self.visualization_code:
                if "distplot" in self.visualization_code:
                    added_codes.append(
                        (
                            [
                                "### Visualization for data distribution of columns",
                            ],
                            "markdown",
                        )
                    )
                    added_codes.append((self.visualization_code["distplot"], "code"))
                if "heatmap" in self.visualization_code:
                    added_codes.append(
                        (
                            [
                                "### Visualization for feature heatmap",
                            ],
                            "markdown",
                        )
                    )
                    added_codes.append((self.visualization_code["heatmap"], "code"))
        except Exception as ex:
            self.logger.warning(f"Warning 134: visualization processing issue ({ex})")
            return []
        return added_codes


class ExecuteNotebookThread(Thread):
    def __init__(self, nb: NotebookNode, resources=None) -> None:
        self.ep = ExecutePreprocessor(timeout=6000, kernel_name="python3")
        self.nb = nb
        self.resources = resources
        self.exception = None
        Thread.__init__(self)

    def run(self):
        try:
            self.ep.preprocess(self.nb, self.resources)
        except Exception as e:
            self.exception = e

    def trigger_interrupt_kernel(self):
        assert self.ep.km
        self.ep.km.interrupt_kernel()

    def get_exception(self):
        return self.exception


class Miner:
    def __init__(
        self,
        folder_path,
        init_blocks=[],
        visualization_code=None,
        logger=None,
        skeleton: Optional[dict] = None,
        explanation: Optional[dict] = None,
        run_info: Optional[dict] = None,
    ):
        self.files = []
        self.content = []
        self.blocks = []
        self.templates = []
        self.visualization_code = visualization_code
        self.logger = logger
        if self.logger is None:
            self.logger = setup_logger()

        self.pipeline_explnation = None
        explains = None

        if skeleton is None:
            self.logger.warning("No Skeleton Information")
        if explanation is None:
            self.logger.warning("No Explanation Information")

        if skeleton and explanation:
            self.pipeline_explnation = Pipeline_Explanation(
                skeleton=skeleton, explanation=explanation, run_info=run_info
            )
            explains = self.pipeline_explnation.process()

        if os.path.isfile(folder_path) and folder_path.endswith("py"):
            # single file processing
            self.files.append(folder_path)
            dir_path = Path(folder_path).parent.absolute().as_posix()
        else:
            # batch processing
            self.files = glob(os.path.join(folder_path, "*.py"))
            dir_path = Path(folder_path).absolute().as_posix()

        self.output_path = dir_path
        self.local_path = dir_path

        for path in self.files:
            with open(path, "r", encoding="utf-8") as f:
                code = f.read()

            code = "\n".join(code.split("\n")[2:])

            current_block = self.get_block(code)
            self.blocks.append(current_block)
            updated_block = self.add_template_block(current_block, self.visualization_code)
            self.content.append(
                {
                    "path": path,
                    "filename": "".join(_ for _ in (path.split("/")[-1].split(".")[:-1])),
                    "code": code,
                    "blocks": updated_block,
                }
            )

        self.templates = {"cells": [], "metadata": {}, "nbformat": 4, "nbformat_minor": 4}

        for index, rec in enumerate(self.content):
            jnote = self.templates
            for block in init_blocks:
                # adding initial_blocks if provided (i.e., from AutoEDA)
                cell_rows, cell_type = block
                if cell_type == "markdown":
                    cell = {"cell_type": "markdown", "metadata": {}, "source": "\n".join(_ for _ in cell_rows)}
                else:
                    # process code type cells
                    cell = {
                        "cell_type": "code",
                        "execution_count": None,
                        "metadata": {},
                        "outputs": [],
                        "source": "\n".join(_ for _ in cell_rows),
                    }
                jnote["cells"].append(cell)

            for block in rec["blocks"]:
                # processing generated blocks
                cell_rows, cell_type = block
                if cell_type == "markdown":
                    cell = {"cell_type": "markdown", "metadata": {}, "source": "\n".join(_ for _ in cell_rows)}
                else:
                    if explains:
                        # code cell
                        # TODO: optimize it or use a tag
                        explain_cell = None
                        found_counter = 0
                        for block_loc in block[0]:
                            if block_loc.strip():
                                for iexplain, explain in enumerate(explains):
                                    if not explain["added"]:
                                        for ex_code in explain["code"]:
                                            for ex_loc in ex_code.split("\n"):
                                                if (
                                                    not (ex_loc.startswith("from ") or ex_loc.startswith("import "))
                                                    and ex_loc.strip()
                                                    and ex_loc in block_loc
                                                ):
                                                    if found_counter == 0:
                                                        found_counter += 1  # at least two matched loc needed
                                                    else:
                                                        explain_cell = {
                                                            "cell_type": "markdown",
                                                            "metadata": {},
                                                            "source": explain["explanation"],
                                                        }
                                                        explains[iexplain]["added"] = True
                                                        found_counter = 0
                                                        break
                                    if explain_cell:
                                        break
                                if explain_cell:
                                    break
                        if explain_cell:
                            jnote["cells"].append(explain_cell)

                    cell = {
                        "cell_type": "code",
                        "execution_count": None,
                        "metadata": {},
                        "outputs": [],
                        "source": "\n".join(_ for _ in cell_rows),
                    }

                jnote["cells"].append(cell)
            rec["jupyter"] = jnote

    @staticmethod
    def __save__(filename, notebook_object):
        with open(filename, "w", encoding="utf-8") as fw:
            json.dump(notebook_object, fw)

    @staticmethod
    def execute_notebook(
        nb: NotebookNode, resources=None, timeout: int = 0, cancel: Optional[CancellationToken] = None
    ):
        ep_thread = ExecuteNotebookThread(nb, resources)
        ep_thread.start()

        start_time = time.time()

        is_interrupted = False
        while not is_interrupted and ep_thread.is_alive():
            if (timeout > 0) and (time.time() - start_time) > timeout:
                ep_thread.trigger_interrupt_kernel()
                is_interrupted = True
            if cancel and cancel.is_triggered:
                ep_thread.trigger_interrupt_kernel()
                is_interrupted = True
            time.sleep(1)
        ep_thread.join()
        exception = ep_thread.get_exception()
        if exception:
            raise exception

    def save_all(self, execution=False, timeout: int = 0, cancel: Optional[CancellationToken] = None):
        output_files = []
        for rec in self.content:
            path = os.path.join(self.output_path, rec["filename"] + ".ipynb")
            self.__save__(path, rec["jupyter"])
            self.logger.info(f"saved:{path}")
            if execution and (cancel is None or cancel.is_triggered is False):
                try:
                    self.logger.info("Running the explained notebook...")
                    with open(path, "r", encoding="utf-8") as f:
                        nb = nbformat.read(f, as_version=4)
                    self.execute_notebook(nb, {"metadata": {"path": self.output_path}}, timeout, cancel)
                except Exception:
                    import traceback

                    self.logger.warning(f"Failed to execute notebook: {path}.ipynb, error: {traceback.format_exc()}")
                finally:
                    with open(f"{path}.out.ipynb", "w", encoding="utf-8") as f:
                        nbformat.write(nb, f)
                    output_files.append(f"{path}.out.ipynb")
                    self.logger.info(f"Saved explained notebook in: {path}.out.ipynb")
        return output_files

    def get_block(self, code):
        """generate a set of blocks from LoCs
        each block seprated through empty lines (\n)
        """
        blocks = []
        current_block = []
        eob = False
        for line in code.split("\n"):
            if len(line.strip()) == 0:
                eob = True
            else:
                if (line.startswith("\t") or line.startswith(" ")) and eob:
                    eob = False
                if (line.startswith("#") or line.startswith("'''")) and eob:
                    # generate a new block
                    blocks.append(current_block)
                    current_block = []
                    eob = False
                current_block.append(line)
        if len(blocks):
            blocks.append(current_block)
        return blocks

    def add_template_block(
        self, blocks, visualization_code, content_template_path=os.path.join("templates", "jupyter_content.json")
    ):
        ct = Code_Template()
        desc = AST_Update(visualization_code, local_path=self.local_path)
        final_block = []
        try:
            path = Path(__file__).parent / content_template_path
            jupyter_content_template = json.load(open(path, "r", encoding="utf-8"))
        except Exception:
            self.logger.error(f"Template file not found:{content_template_path}")
            raise
        prev = ""
        for block in blocks:
            before_block = []
            after_block = []
            extended_lines = []
            for line in block:
                if line.strip() in jupyter_content_template:
                    before_block, after_block = jupyter_content_template[line.strip()]
                try:
                    desc_lines = desc.process(line, prev)
                    if desc_lines:
                        extended_lines.extend(desc_lines)
                except NotImplementedError:
                    raise
                prev = line

            if before_block:
                before_block = ct.update(before_block)
                final_block.append((before_block, "markdown"))
            if extended_lines:
                final_block.extend(extended_lines)

            final_block.append((block, "code"))

            if after_block:
                after_block = ct.update(after_block)
                final_block.append((after_block, "markdown"))

        return final_block



# Copyright 2023 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
from sapientml.util.logging import setup_logger
from scipy import stats


class Dataset:
    def __init__(self, dataframe: pd.DataFrame, target_columns: list[str]):
        self.data = dataframe
        self.target = target_columns


class EDA(Dataset):
    def __init__(self, dataframe: pd.DataFrame, target_columns: list[str], ref="Dataset", log_level=1, logger=None):
        self.logger = logger
        if logger is None:
            self.logger = setup_logger()
        self.logger.info("AutoEDA is processing input dataset...")
        self.df = Dataset(dataframe, target_columns)
        self.log_level = log_level
        self.description = []
        self.ref = ref
        self.add_general_block()
        self.check_skewness()

    def __str__(self):
        lines = ""
        for i, j in self.description:
            comm = "#" if j == "markdown" else ""
            lines += comm + "\n".join(_ for _ in i) + "\n"
        return lines

    def check_skewness(self):
        ref = self.ref
        skews = []
        for col in self.df.data:
            try:
                skew = self.df.data[col].skew()
            except TypeError:
                skew = 0
            if skew > 1 or skew < -1:
                skews.append([col, skew, "highly skewed"])
            elif skew > 0.5 or skew < -0.5:
                skews.append([col, skew, "moderately skewed"])

        if len(skews):
            skews = pd.DataFrame(skews, columns=["Column", "Skewness", "Skewness_Type"])

            self.description.append(
                (
                    [
                        "## Skewness",
                        "In probability theory and statistics, skewness is a measure of the asymmetry of the probability distribution of a real-valued random variable about its mean. The skewness value can be positive, zero, negative, or undefined.",
                        "\nMore detail can be found [here](https://en.wikipedia.org/wiki/Skewness) and the [Probability and Statistics Tables and Formulae](http://tomlr.free.fr/Math%E9matiques/Math%20Complete/Probability%20and%20statistics/CRC%20-%20standard%20probability%20and%20Statistics%20tables%20and%20formulae%20-%20DANIEL%20ZWILLINGER.pdf) by Zwillinger and Kokoska.",
                        '\nHere are two samples Skewness data for positive and negative skew data.<img src="https://upload.wikimedia.org/wikipedia/commons/f/f8/Negative_and_positive_skew_diagrams_%28English%29.svg" alt="source:Wikimedia">',
                        f"\nFirst, we will calculate the skewness for each column in {ref} if each selected column has positive float/integer values.",
                        "\nSecond, we will review the following conditions.",
                        "\n- if $Skewness>1$ or $Skewness<-1$, we will consider it as highly skewed;",
                        "\n- if $0.5<Skewness<=1$ or $-0.5<Skewness<=-1$, we will consider it as moderate skewed;",
                        f"\n{skews.to_html()}",
                    ],
                    "markdown",
                )
            )

    def add_general_block(self):
        ref = self.ref
        if self.df.data is not None:
            n_rows = self.df.data.shape[0]
            n_col = self.df.data.shape[1]
            n_types = self.df.data.dtypes.unique()
            self.description.append(
                (
                    [
                        "# Exploratory Data Analysis (EDA)",
                        "## General Structure",
                        f"{ref} includes **{n_col}** columns and **{n_rows}** rows.",
                        f"There are **{len(n_types)}** different data types as follows: "
                        + f"*{u', '.join(str(_) for _ in n_types.tolist())}*.",
                    ],
                    "markdown",
                )
            )
            self.description.append(
                (
                    [
                        "Let's review the dataset description:",
                        f"{self.df.data.describe().T.to_html()}",
                    ],
                    "markdown",
                )
            )
            all_sum = self.df.data.isnull().sum()
            all_sum = all_sum[all_sum != 0]

            if len(all_sum):
                # generate content if there is a null value
                all_sum.sort_values(ascending=False, inplace=True)
                all_sum_filter = pd.DataFrame(
                    [(str(all_sum.index[i]), str(u)) for i, u in enumerate(all_sum[:20])], columns=["Column", "#Null"]
                )
                self.description.append(
                    (
                        [
                            f"**Is there any null value?** \nThe answer is **Yes**; let's review top {'20' if len(all_sum)>20 else str(len(all_sum))} of those columns with the number of Null values.",
                            f"\n{all_sum_filter.to_html()}",
                            f"\nAs partial of the results shown above, there are total **{len(all_sum)}** columns with Null values.",
                        ],
                        "markdown",
                    )
                )

            if self.df.target:
                if isinstance(self.df.target, list):
                    for col in self.df.target:
                        if col in self.df.data:
                            unique_vals = self.df.data[col].unique()
                            if len(unique_vals) < 11:
                                freq = pd.DataFrame(self.df.data[col].value_counts())
                                self.description.append(
                                    (
                                        [
                                            f"There is **{len(unique_vals)}** unique value in **{col}** column which is a target column.",
                                            "let's see frequency of values for the target column of {col}:",
                                            f"{freq.to_html()}",
                                        ],
                                        "markdown",
                                    )
                                )

    def cat_process(self, threshold=0.01, IQR_activation: bool = True, z_activation: bool = True):
        hashmaps = {}
        desc = []
        ratio = float(threshold / 100)
        for key in self.df.data:
            hashmaps = dict(self.df.data.pivot_table(columns=[key], aggfunc="size"))
            # skip all unique values &
            # skip all keys with single repatation (unique) &
            # skip large keys (100+) &
            # skip none values &
            if hashmaps:
                if (
                    len(hashmaps.keys()) != self.df.data[key].shape[0]
                    and sum(hashmaps.values()) != len(hashmaps.values())
                    and 1 < len(hashmaps.keys()) < 100
                    and len(hashmaps.values())
                ):
                    values = list(hashmaps.values())

                    # baseline
                    upper = np.quantile(values, 1 - ratio)
                    lower = np.quantile(values, ratio)

                    # IQR
                    Q1 = np.quantile(values, 0.25)
                    Q3 = np.quantile(values, 0.75)
                    IQR = Q3 - Q1

                    # Z-Score
                    z = np.abs(stats.zscore(values))

                    for index, cat in enumerate(hashmaps):
                        val = hashmaps[cat]
                        if (
                            val > upper
                            and (not IQR_activation or val > (Q3 + 1.5 * IQR))
                            and (not z_activation or z[index] > 3)
                        ):
                            desc.append((key, cat, hashmaps[cat], lower, upper, "Upper"))
                        elif (
                            val < lower
                            and (not IQR_activation or val < (Q1 - 1.5 * IQR))
                            and (not z_activation or z[index] > 3)  # we already have abs of z-score
                        ):
                            desc.append((key, cat, hashmaps[cat], lower, upper, "Lower"))
        df_desc = pd.DataFrame(desc, columns=["Field", "Value", "Frequency", "Lower", "Upper", "Criteria"])
        if len(df_desc):
            singular = "s" if len(df_desc) > 1 else ""
            singular_verb = "are" if len(df_desc) > 1 else "is"
            exp_statement = "> Upper(C_0)" if df_desc.iloc[0].Criteria == "Upper" else "< Lower(C_0)"
            self.description.append(
                (
                    [
                        "# Finding Intresting Datapoints",
                        "Let's process each field by their histogram frequency and check if there is any intresting data point.",
                        f"\nThere {singular_verb} **{len(df_desc)}** number of intresting value{singular} in the following column{singular}.",
                        "The below table shows each **Value** of each **Field**(column) with their total frequencies, **Lower** shows the lower frequency of normal distribution, **Upper** shows the upper bound frequency of normal distribution, and **Criteria** shows if the frequnecy passed **Upper bound** or **Lower bound**.",
                        f"{df_desc.to_html()}",
                        "\n\n",
                        f"For example, in the **{df_desc.iloc[0].Field}** column the value of **{df_desc.iloc[0].Value}** has **{df_desc.iloc[0].Frequency}** repeatation but this number is not between Lower bound({df_desc.iloc[0].Lower}) and Upper bound({df_desc.iloc[0].Upper}).\n\n",
                        f"Let     $C_0={df_desc.iloc[0].Value}$   and   $Freq(C_0)={df_desc.iloc[0].Frequency}$     ,   $Upper(C_0)={df_desc.iloc[0].Upper}$     ,   $Lower(C_0)={df_desc.iloc[0].Lower}$\n",
                        f"$Freq(C_0) {exp_statement}$.",
                    ],
                    "markdown",
                )
            )
        return hashmaps, df_desc

    def check_consistency(self, convert=False):
        for key in self.df.data:
            if self.df.data[key].dtype == "object":
                self.df.data[key] = self.__convert_object(self.df.data[key], convert, key)
        return self.df.data

    def __is_bool(self, unique_vals):
        """check if the given input can be considered as boolean"""
        if len(unique_vals) != 2:
            # total number of elements sould be two, to be considered
            return None
        sorted_unique_vals = sorted([_.lower().strip() for _ in unique_vals])
        candidates = [
            # TODO: add dynamic values
            ["no", "yes"],
            ["false", "true"],
            ["incorrect", "correct"],
            ["0", "1"],
            ["n", "y"],
            ["f", "t"],
        ]

        for candidate in candidates:
            if sorted(candidate) == sorted_unique_vals:
                # return consistant index (i.e., yes=1 and no=0)
                return [candidate.index(_) for _ in sorted_unique_vals]
        return None

    def __convert_digit(self, cell, key):
        ref = self.ref
        try:
            cell = cell.astype(int)
        except ValueError:
            # if there is float value in list
            cell = cell.astype(float)
            if self.log_level >= 2:
                self.description.append(
                    (
                        [
                            f"Let's convert {ref}['{key}'] from Object to float because it includes at least a float values."
                        ],
                        "markdown",
                    )
                )
            self.description.append(([f"{ref}['{key}']={ref}['{key}'].astype(float)"], "code"))
        else:
            if self.log_level >= 2:
                self.description.append(
                    (
                        [f"Let's convert {ref}['{key}'] from Object to integer."],
                        "markdown",
                    )
                )
            self.description.append(([f"{ref}['{key}']={ref}['{key}'].astype(int)"], "code"))
        return cell

    def __convert_object(self, cell, convert: bool, key):
        """
        types source: https://pandas.pydata.org/docs/reference/api/pandas.Series.str.isnumeric.html
        :params convert: convert bool string to numerical val
        :params key: the current title of column
        """
        ref = self.ref

        numeric = cell[pd.to_numeric(cell, errors="coerce").notnull()]
        #        alpha=cell[cell.str.isalpha()]
        alpha = cell[(cell.str.isalpha()) | (cell.isna())]

        if len(alpha) == 0 and len(numeric) == len(cell):  # object but no alpha  # mixed of int/float
            return self.__convert_digit(cell, key) if convert else cell
        elif len(alpha) > 0:  # object: mixed of alpha and possible int/float
            unique_num = numeric.unique().tolist()
            unique_alpha = alpha.unique().tolist()

            bool_check = self.__is_bool(unique_alpha)

            if bool_check is not None and len(unique_num) > len(  # there is bool transformer
                unique_alpha
            ):  # numerical category is bigger
                # we have boolean values
                # replace values with bool_candidate indexes
                if self.log_level >= 1:
                    self.description.append(
                        (
                            [
                                f"{ref}['{key}'] includes string values of: {','.join(_ for _ in unique_alpha)}  where it is mixed with {len(unique_num)} different number of numeric values."
                            ],
                            "markdown",
                        )
                    )
                if convert:
                    for index, val in enumerate(unique_alpha):
                        cell = cell.replace(val, str(bool_check[index]))
                        if self.log_level >= 2:
                            self.description.append(
                                (
                                    [f"We can replace '{val}' with '{str(bool_check[index])}'"],
                                    "markdown",
                                )
                            )
                        self.description.append(
                            ([f"{ref}['{key}']={ref}['{key}'].replace('{val}',str({bool_check[index]}))"], "code")
                        )
                    # convert to target int or float
                    cell = self.__convert_digit(cell, key)
        return cell



# Copyright 2023 The SapientML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import textwrap
from pathlib import Path

from jinja2 import Environment, FileSystemLoader
from pydantic import BaseModel
from sapientml import macros
from sapientml.util.json_util import JSONEncoder
from sapientml.util.logging import setup_logger

from ...params import Pipeline
from ...seeding.predictor import name_to_label_mapping

logger = setup_logger()
env = Environment(loader=FileSystemLoader(f"{os.path.dirname(__file__)}/../../templates"), trim_blocks=True)

MODEL_IMPORT_LIBRARY_MAP = {
    "XGBRegressor": "xgboost",
    "XGBClassifier": "xgboost",
    "RandomForestRegressor": "sklearn.ensemble",
    "RandomForestClassifier": "sklearn.ensemble",
    "LogisticRegression": "sklearn.linear_model",
    "LinearRegression": "sklearn.linear_model",
    "LinearSVC": "sklearn.svm",
    "LinearSVR": "sklearn.svm",
    "KNeighborsClassifier": "sklearn.neighbors",
    "GradientBoostingRegressor": "sklearn.ensemble",
    "GradientBoostingClassifier": "sklearn.ensemble",
    "ExtraTreesRegressor": "sklearn.ensemble",
    "ExtraTreesClassifier": "sklearn.ensemble",
    "CatBoostRegressor": "catboost",
    "CatBoostClassifier": "catboost",
    "DecisionTreeRegressor": "sklearn.tree",
    "DecisionTreeClassifier": "sklearn.tree",
    "SGDRegressor": "sklearn.linear_model",
    "SGDClassifier": "sklearn.linear_model",
    "LGBMRegressor": "lightgbm",
    "LGBMClassifier": "lightgbm",
    "MLPClassifier": "sklearn.neural_network",
    "MLPRegressor": "sklearn.neural_network",
    "MultinomialNB": "sklearn.naive_bayes",
    "BernoulliNB": "sklearn.naive_bayes",
    "SVC": "sklearn.svm",
    "SVR": "sklearn.svm",
    "AdaBoostClassifier": "sklearn.ensemble",
    "AdaBoostRegressor": "sklearn.ensemble",
    "Lasso": "sklearn.linear_model",
    "GaussianNB": "sklearn.naive_bayes",
}

NO_RANDOM_SEED_MODELS = [
    "SVR",
    "LinearRegression",
    "MultinomialNB",
    "GaussianNB",
    "BernoulliNB",
]

NO_TUNABLE_PARAMS_MODELS = [
    "LinearRegression",
    "MultinomialNB",
    "GaussianNB",
    "BernoulliNB",
]


def load_json(file_name):
    with open(file_name, "r", encoding="utf-8") as input_file:
        content = json.load(input_file)
    return content


def is_allowed_to_apply_to_target(label_name: str) -> bool:
    return "PREPROCESS:Scaling:log" in label_name


class PipelineTemplate(BaseModel):
    pipeline: Pipeline

    def _render(self, tpl, *args, **kwargs):
        code = tpl.render(*args, **kwargs)
        return "\n".join([line for line in code.split("\n") if len(line) > 0])

    def generate(self):
        pipeline = self.pipeline
        if pipeline.model is None:
            return
        model_name = pipeline.model.label_name.split(":")[2]
        is_boolean_target = any(
            [pipeline.dataset_summary.columns[col].dtype == "bool" for col in pipeline.task.target_columns]
        )
        if model_name == "CatBoostClassifier" and is_boolean_target:
            target2string = True
        else:
            target2string = False

        # Use tpl.render but self._render because keep blank lines.
        tpl = env.get_template("other_templates/store_prediction_columns.py.jinja")
        pipeline.pipeline_json["store_prediction_columns"]["code"] = self._render(tpl, pipeline=pipeline)

        if len(pipeline.task.ignore_columns) > 0:
            pipeline.train_ignore_columns = pipeline.task.ignore_columns
            pipeline.test_ignore_columns = pipeline.task.ignore_columns
            tpl = env.get_template("other_templates/drop_columns.py.jinja")
            pipeline.pipeline_json["discard_columns"]["code"] = self._render(tpl, pipeline=pipeline)
            tpl = env.get_template("other_templates/drop_columns_train.py.jinja")
            pipeline.pipeline_json["discard_columns"]["code_train"] = self._render(tpl, pipeline=pipeline)
            tpl = env.get_template("other_templates/drop_columns_predict.py.jinja")
            pipeline.pipeline_json["discard_columns"]["code_predict"] = self._render(tpl, pipeline=pipeline)

        tpl = env.get_template("other_templates/target_separation_validation.py.jinja")
        pipeline.pipeline_json["target_separation"]["code_validation"] = self._render(tpl, pipeline=pipeline)
        tpl = env.get_template("other_templates/target_separation_test.py.jinja")
        pipeline.pipeline_json["target_separation"]["code_test"] = self._render(tpl, pipeline=pipeline)
        tpl = env.get_template("other_templates/target_separation_train.py.jinja")
        pipeline.pipeline_json["target_separation"]["code_train"] = self._render(tpl, pipeline=pipeline)
        tpl = env.get_template("other_templates/target_separation_predict.py.jinja")
        pipeline.pipeline_json["target_separation"]["code_predict"] = self._render(tpl, pipeline=pipeline)

        if pipeline.inverse_target:
            tpl = env.get_template("other_templates/inverse_target.py.jinja")
            pipeline.pipeline_json["inverse_target"]["code"] = self._render(tpl, pipeline=pipeline)

        tpl = env.get_template("other_templates/evaluation.py.jinja")
        code = self._render(tpl, pipeline=pipeline, target2string=target2string, macros=macros)
        pipeline.pipeline_json["evaluation"]["code_validation"] = code
        pipeline.pipeline_json["evaluation"]["code_test"] = textwrap.indent(code, "    ")
        tpl = env.get_template("other_templates/prediction_result.py.jinja")
        pipeline.pipeline_json["output_prediction"]["code"] = self._render(tpl, pipeline=pipeline, macros=macros)

        flag_hyperparameter_tuning = (
            pipeline.config.hyperparameter_tuning and model_name not in NO_TUNABLE_PARAMS_MODELS
        )
        if flag_hyperparameter_tuning:
            tpl = env.get_template("model_templates/hyperparameters.py.jinja")
            pipeline.pipeline_json["hyperparameters"]["code"] = self._render(tpl, model_name=model_name)

            tpl = env.get_template("model_templates/hyperparameters_default_value.py.jinja")
            pipeline.pipeline_json["hyperparameters_default_value"]["code"] = self._render(tpl, model_name=model_name)

            tpl = env.get_template("other_templates/hyperparameter_tuning_evaluation.py.jinja")
            pipeline.pipeline_json["hyperparameter_tuning_evaluation"]["code"] = self._render(
                tpl, pipeline=pipeline, target2string=target2string, macros=macros
            )

            self.populate_hyperparameter_tuning()
        elif pipeline.config.hyperparameter_tuning and model_name in NO_TUNABLE_PARAMS_MODELS:
            logger.debug(
                "'hyperparameter_tuning' is set to True, but the candidate model("
                + model_name
                + ") doesn't have tunable parameters."
            )

        self.populate_model()

        if pipeline.config.permutation_importance:
            tpl = env.get_template("other_templates/permutation_importance.py.jinja")
            pipeline.pipeline_json["permutation_importance"]["code"] = self._render(
                tpl, pipeline=pipeline, target2string=target2string
            )

        tpl_validation = env.get_template("pipeline_validation.py.jinja")
        pipeline.validation = tpl_validation.render(
            pipeline_json=pipeline.pipeline_json,
            pipeline=pipeline,
            flag_hyperparameter_tuning=flag_hyperparameter_tuning,
        )
        tpl_test = env.get_template("pipeline_test.py.jinja")
        pipeline.test = tpl_test.render(
            pipeline_json=pipeline.pipeline_json,
            pipeline=pipeline,
            flag_hyperparameter_tuning=flag_hyperparameter_tuning,
        )

        tpl_train = env.get_template("pipeline_train.py.jinja")
        pipeline.train = tpl_train.render(
            pipeline_json=pipeline.pipeline_json,
            pipeline=pipeline,
            flag_hyperparameter_tuning=flag_hyperparameter_tuning,
        )
        tpl_predict = env.get_template("pipeline_predict.py.jinja")
        pipeline.predict = tpl_predict.render(
            pipeline_json=pipeline.pipeline_json,
            pipeline=pipeline,
            flag_hyperparameter_tuning=flag_hyperparameter_tuning,
        )

    def add_processing_components(self, component, type, training_dataframe, test_dataframe):
        snippets = self.generate_snippet(component, training_dataframe, test_dataframe)
        snippets_train = self.generate_snippet(component, training_dataframe, test_dataframe, "train")
        snippets_predict = self.generate_snippet(component, training_dataframe, test_dataframe, "predict")

        if len(snippets) > 0:
            # prepend explanation to the snippet of the component; first snippet in case of multiple snippets
            component_json, explanation = self.create_preprocessing_component_explanation(component)
            snippets[0] = f"{explanation}\n{snippets[0]}"

            component_json["code"] = snippets.copy()
            component_json["code_train"] = snippets_train.copy()
            component_json["code_predict"] = snippets_predict.copy()
            self.pipeline.pipeline_json[type][component.label_name] = component_json

    def generate_snippet(self, component, training_dataframe, test_dataframe, train_predict_flag=""):
        adapted_snippets = list()

        pipeline = self.pipeline

        relevant_cols = component.relevant_columns
        if not is_allowed_to_apply_to_target(component.label_name):
            relevant_cols = sorted(list(set(relevant_cols) - set(pipeline.task.target_columns)))
        api_label = component.label_name.split(":")[2]

        if len(relevant_cols) == 0 and api_label not in ["STANDARD", "SMOTE"]:
            logger.debug("No relevant columns found for " + api_label)
            return adapted_snippets

        # read component template(s)
        template_files = []
        path = str(Path(os.path.dirname(__file__)) / "../../templates/preprocessing_templates")
        for file in os.listdir(path):
            if file.startswith(api_label):
                if api_label in ["SMOTE"] and train_predict_flag in ["", "train"]:
                    template_files.append(f"preprocessing_templates/{file}")
                elif train_predict_flag == "":
                    if ("train" not in file) and ("predict" not in file):
                        template_files.append(f"preprocessing_templates/{file}")
                elif train_predict_flag != "":
                    if train_predict_flag in file:
                        template_files.append(f"preprocessing_templates/{file}")

        if len(template_files) == 1:
            tpl = env.get_template(template_files[0])
            adapted_snippets.append(
                self._render(
                    tpl,
                    columns=relevant_cols,
                    train_dataset=training_dataframe,
                    test_dataset=test_dataframe,
                    pipeline=pipeline,
                )
            )
        elif len(template_files) > 1:
            # do more analysis to figure out the right template
            if "type" in template_files[0]:
                numeric_cols = []
                other_cols = []

                numeric_dtypes = ["int16", "int32", "int64", "float16", "float32", "float64"]
                for c in relevant_cols:
                    c_dtype = pipeline.all_columns_datatypes[c]
                    if c_dtype in numeric_dtypes:
                        numeric_cols.append(c)
                    else:
                        other_cols.append(c)

                if len(numeric_cols) > 0:
                    template_filename = [x for x in template_files if "numeric" in x]
                    tpl = env.get_template(template_filename[0])
                    cols_almost_missing_numeric = pipeline.dataset_summary.cols_almost_missing_numeric or []
                    adapted_snippets.append(
                        self._render(
                            tpl,
                            cols_almost_missing_numeric=sorted(
                                list(set(numeric_cols) & set(cols_almost_missing_numeric))
                            ),  # because cols_almost_missing_numeric can include columns in ignore_columns, while numeric_cols doesn't include them.
                            columns=sorted(list(set(numeric_cols) - set(cols_almost_missing_numeric))),
                            train_dataset=training_dataframe,
                            test_dataset=test_dataframe,
                        )
                    )
                if len(other_cols) > 0:
                    template_filename = [x for x in template_files if "string" in x]
                    tpl = env.get_template(template_filename[0])
                    cols_almost_missing_string = pipeline.dataset_summary.cols_almost_missing_string or []
                    adapted_snippets.append(
                        self._render(
                            tpl,
                            cols_almost_missing_string=sorted(
                                list(set(other_cols) & set(cols_almost_missing_string))
                            ),  # because cols_almost_missing_string can include columns in ignore_columns, while other_cols doesn't include them.
                            columns=sorted(list(set(other_cols) - set(cols_almost_missing_string))),
                            train_dataset=training_dataframe,
                            test_dataset=test_dataframe,
                        )
                    )
        else:
            logger.debug("ERROR: No template found for component ", api_label)

        return adapted_snippets

    def populate_hyperparameter_tuning(self):
        pipeline = self.pipeline
        if pipeline.model is None:
            return
        model_name = pipeline.model.label_name.split(":")[2]
        # change "predict" to "predict_proba", e.g., for metric = LogLoss, ROC_AUC, Gini since they require probability to be calculated
        if pipeline.adaptation_metric and (
            pipeline.adaptation_metric in macros.metric_needing_predict_proba
            or pipeline.adaptation_metric.startswith(macros.Metric.MAP_K.value)
            or pipeline.config.predict_option == macros.PRED_PROBABILITY
        ):
            tpl = env.get_template("model_templates/classification_post_process.jinja")
            flag_predict_proba = True
            binary_classification_snippet = self._render(tpl, pipeline=pipeline)
        else:
            binary_classification_snippet = ""
            flag_predict_proba = False

        flag_no_random_seed_model = model_name in NO_RANDOM_SEED_MODELS
        _is_multioutput_regression = (
            pipeline.task.task_type == macros.TASK_REGRESSION and len(pipeline.task.target_columns) > 1
        )
        _is_multioutput_classification = (
            pipeline.task.task_type == macros.TASK_CLASSIFICATION and len(pipeline.task.target_columns) > 1
        )

        tpl = env.get_template("model_templates/hyperparameter_tuning.py.jinja")
        snippet = self._render(
            tpl,
            pipeline=pipeline,
            macros=macros,
            import_library=MODEL_IMPORT_LIBRARY_MAP[model_name],
            model_name=model_name,
            params=pipeline.pipeline_json["hyperparameters"]["code"],
            evaluation=pipeline.pipeline_json["hyperparameter_tuning_evaluation"]["code"],
            enqueue_default_hyperparameters=pipeline.pipeline_json["hyperparameters_default_value"]["code"],
            binary_classification_snippet=binary_classification_snippet,
            flag_no_random_seed_model=flag_no_random_seed_model,
            is_multioutput_regression=_is_multioutput_regression,
            is_multioutput_classification=_is_multioutput_classification,
            flag_predict_proba=flag_predict_proba,
            timeout=pipeline.config.hyperparameter_tuning_timeout
            if pipeline.config.hyperparameter_tuning_timeout > 0
            else None,
        )
        pipeline.pipeline_json["hyperparameter_optimization"]["code"] = snippet

    def populate_model(self):
        pipeline = self.pipeline
        if pipeline.model is None:
            return
        model_component_json = dict()
        model_name = pipeline.model.label_name.split(":")[2]
        tpl = env.get_template("model_templates/model.py.jinja")
        tpl_train = env.get_template("model_templates/model_train.py.jinja")
        tpl_predict = env.get_template("model_templates/model_predict.py.jinja")

        flag_no_random_seed_model = model_name in NO_RANDOM_SEED_MODELS
        flag_hyperparameter_tuning = (
            pipeline.config.hyperparameter_tuning and model_name not in NO_TUNABLE_PARAMS_MODELS
        )
        if flag_hyperparameter_tuning and flag_no_random_seed_model:
            model_arg = "HPO_noRandomSeed"
        elif flag_hyperparameter_tuning and not flag_no_random_seed_model:
            model_arg = "HPO_RandomSeed"
        elif not flag_hyperparameter_tuning and flag_no_random_seed_model:
            model_arg = "noHPO_noRandomSeed"
        else:
            model_arg = "noHPO_RandomSeed"
        flag_predict_proba = (
            (pipeline.adaptation_metric in macros.metric_needing_predict_proba)
            or (pipeline.adaptation_metric.startswith(macros.Metric.MAP_K.value))
            or (pipeline.config.predict_option == macros.PRED_PROBABILITY)
        )
        _is_multioutput_regression = (
            pipeline.task.task_type == macros.TASK_REGRESSION and len(pipeline.task.target_columns) > 1
        )
        _is_multioutput_classification = (
            pipeline.task.task_type == macros.TASK_CLASSIFICATION and len(pipeline.task.target_columns) > 1
        )
        snippet = self._render(
            tpl,
            pipeline=pipeline,
            import_library=MODEL_IMPORT_LIBRARY_MAP[model_name],
            model_name=model_name,
            params=pipeline.model.hyperparameters or "",
            model_arg=model_arg,
            flag_predict_proba=flag_predict_proba,
            is_multioutput_regression=_is_multioutput_regression,
            is_multioutput_classification=_is_multioutput_classification,
            metric_needing_predict_proba=macros.metric_needing_predict_proba,
        )
        snippet_train = self._render(
            tpl_train,
            pipeline=pipeline,
            import_library=MODEL_IMPORT_LIBRARY_MAP[model_name],
            model_name=model_name,
            params=pipeline.model.hyperparameters or "",
            model_arg=model_arg,
            flag_predict_proba=flag_predict_proba,
            is_multioutput_regression=_is_multioutput_regression,
            is_multioutput_classification=_is_multioutput_classification,
            metric_needing_predict_proba=macros.metric_needing_predict_proba,
        )
        snippet_predict = self._render(
            tpl_predict,
            pipeline=pipeline,
            import_library=MODEL_IMPORT_LIBRARY_MAP[model_name],
            model_name=model_name,
            params=pipeline.model.hyperparameters or "",
            model_arg=model_arg,
            flag_predict_proba=flag_predict_proba,
            is_multioutput_regression=_is_multioutput_regression,
            is_multioutput_classification=_is_multioutput_classification,
            metric_needing_predict_proba=macros.metric_needing_predict_proba,
        )

        # change "predict" to "predict_proba", e.g., for metric = LogLoss, ROC_AUC, Gini since they require probability to be calculated
        if pipeline.adaptation_metric and (
            pipeline.adaptation_metric in macros.metric_needing_predict_proba
            or pipeline.adaptation_metric.startswith(macros.Metric.MAP_K.value)
            or pipeline.config.predict_option == macros.PRED_PROBABILITY
        ):
            snippet = snippet.replace("predict", "predict_proba")
            tpl = env.get_template("model_templates/classification_post_process.jinja")
            snippet += "\n" + self._render(tpl, pipeline=pipeline)

            snippet_predict = snippet_predict.replace("predict", "predict_proba")
            tpl_predict = env.get_template("model_templates/classification_post_process.jinja")
            snippet_predict += "\n" + self._render(tpl_predict, pipeline=pipeline)

        model_component_json["code"] = snippet
        model_component_json["code_train"] = snippet_train
        model_component_json["code_predict"] = snippet_predict

        with open(Path(os.path.dirname(__file__)) / "../../models/feature_importance.json", "r", encoding="utf-8") as f:
            model_feature_weights = json.load(f)

        for key, task_label in name_to_label_mapping.items():
            if pipeline.model.label_name in task_label.values():
                model_name = key
                break

        pipeline.model.meta_features = (
            model_feature_weights[model_name][:5] if model_name in model_feature_weights else []
        )

        # prepend explanation to the model snippet
        snippet = self.create_model_component_explanation(pipeline.model, model_component_json) + snippet
        pipeline.pipeline_json["model"] = model_component_json
        return snippet

    def create_preprocessing_component_explanation(self, component):
        component_description_dict = load_json(
            str(
                Path(os.path.dirname(__file__))
                / "../../templates/explainability_templates"
                / "component_description.json"
            )
        )
        entry = component_description_dict[component.label_name]
        api_description = entry["api_description"]
        data_shape = entry["data_shape"]
        hyperparameters = entry["hyperparameters"]
        relevant_metafeatures = [o.feature_name for o in component.predicate_objects]

        # special case for standard scaler, since it is applied to the entire dataset
        if (
            component.label_name == "PREPROCESS:Scaling:STANDARD:sklearn"
            or component.label_name == "PREPROCESS:Balancing:SMOTE:imblearn"
        ):
            component.relevant_columns = "all columns in the dataset"

        ctx = dict()
        ctx["target_component_name"] = component.label_name
        ctx["relevant_meta_feature_list"] = relevant_metafeatures
        ctx["relevant_column_list"] = component.relevant_columns
        ctx["api_description"] = api_description
        ctx["data_shape"] = data_shape

        if len(component.components_before) > 0:
            ctx["before_or_after"] = "after"
            ctx["dependent_component_list"] = component.components_before
        elif len(component.components_after) > 0:
            ctx["before_or_after"] = "before"
            ctx["dependent_component_list"] = component.components_after

        if len(component.alternative_components) > 0:
            ctx["alternative_component_list"] = component.alternative_components
            alt_comp_predicates = list()
            comp_predicates = [o.feature_name for o in component.predicate_objects]
            for alt_comp in component.alternative_components:
                alt_comp_predicates.extend([o.feature_name for o in alt_comp.predicate_objects])

            metafeature_1 = ""
            metafeature_2 = ""
            for p, alt_p in zip(comp_predicates, alt_comp_predicates):
                if p == alt_p:
                    continue
                else:
                    metafeature_1 = p
                    metafeature_2 = alt_p
                    break
            ctx["relevant_meta_feature_1"] = metafeature_1
            ctx["relevant_meta_feature_2"] = metafeature_2

        hp_text = ""
        if len(hyperparameters) > 0:
            for hp in hyperparameters:
                hp_text += '\n#\t\t "' + hp["hp_name"] + ": " + hp["hp_values"] + '" :: ' + hp["hp_description"]
        else:
            hp_text = "None"
        ctx["hyperparameters_description"] = hp_text

        tpl = env.get_template("explainability_templates/preprocessing_explanation.py.jinja")
        explanation = self._render(tpl, **ctx)

        # normalize component names
        for old_name, map in component_description_dict.items():
            explanation = explanation.replace(old_name, map["normalized_name"])

        # populate json
        component_json = dict()
        component_json["explanation"] = dict()
        component_json["explanation"]["target_component_name"] = component.label_name
        component_json["explanation"]["relevant_meta_feature_list"] = relevant_metafeatures
        component_json["explanation"]["relevant_column_list"] = component.relevant_columns
        component_json["explanation"]["api_description"] = api_description
        component_json["explanation"]["data_shape"] = data_shape
        component_json["explanation"]["hyperparameters_description"] = hyperparameters
        component_json["explanation"]["alternative_component_list"] = list()
        for alt_comp in component.alternative_components:
            component_json["explanation"]["alternative_component_list"].append(
                {
                    "component_name": alt_comp.label_name,
                    "relevant_meta_feature_list": [o.feature_name for o in alt_comp.predicate_objects],
                }
            )
        component_json["explanation"]["before_dependent_component_list"] = component.components_before
        component_json["explanation"]["after_dependent_component_list"] = component.components_after

        return component_json, explanation

    def create_model_component_explanation(self, component, model_component_json):
        tpl = env.get_template("explainability_templates/model_explanation.py.jinja")
        explanation = self._render(
            tpl, target_component_name=component.label_name, relevant_meta_feature_list=component.meta_features
        )

        # populate json
        model_component_json["explanation"] = dict()
        model_component_json["explanation"]["target_component_name"] = component.label_name
        model_component_json["explanation"]["relevant_meta_feature_list"] = component.meta_features

        return explanation

    def save_pipeline_json(self):
        # json_string = json.dumps(self.pipeline_json)
        # with open(self.output_dir_path + 'pipeline_json.json', 'w') as outfile:
        #     json.dump(json_string, outfile)

        with open(self.pipeline.output_dir_path + "pipeline_json.json", "w", encoding="utf-8") as f:
            json.dump(self.pipeline.pipeline_json, f, cls=JSONEncoder, indent=4)
